"{ Package: 'jv:vdb' }"

Stream subclass:#VDBInternalPipeStream
	instanceVariableNames:'buffer first last dataAvailable spaceAvailable closed'
	classVariableNames:'DefaultBufferSize'
	poolDictionaries:''
	category:'VDB-Support'
!

!VDBInternalPipeStream class methodsFor:'documentation'!

documentation
"
    not useful on its own, but can be used to talk to a vt100
    terminal view ...
    See example.
"
!

examples
"
                                                                [exBegin]
    |p|

    p := InternalPipeStream new.
    [
        10 timesRepeat:[
            p nextPutLine:'hello'
        ].
    ] fork.

    [
        10 timesRepeat:[
            Transcript showCR:p nextLine
        ].
    ] fork.
                                                                [exEnd]

                                                                [exBegin]
    |userInput elizasOutput top terminal|

    userInput    := InternalPipeStream new.
    elizasOutput := InternalPipeStream new.

    top := StandardSystemView new.
    terminal := VT100TerminalView openOnInput: userInput output:elizasOutput in:top.

    top extent:(terminal preferredExtent).
    top label:'The doctor is in'.
    top iconLabel:'doctor'.
    top open.
    top waitUntilVisible.

    terminal translateNLToCRNL:true.
    terminal inputTranslateCRToNL:true.
    terminal localEcho:true.

    elizasOutput nextPutLine:'Hi, I am Eliza'.
    elizasOutput nextPutLine:'What is your problem ?'.
    elizasOutput nextPutLine:''.
    elizasOutput nextPutAll:'>'.

    [top realized] whileTrue:[
        |line answer matchingRule|

        line := userInput nextLine.
        (#('quit' 'exit' 'end' 'bye') includes:line) ifTrue:[
            top destroy.
            ^ self
        ].

        answer := 'Tell me more.'.
        elizasOutput nextPutLine:answer.
        elizasOutput nextPutAll:'>'.
    ].
                                                                [exEnd]
"
! !

!VDBInternalPipeStream class methodsFor:'initialization'!

initialize
    "Invoked at system start or when the class is dynamically loaded."

    "/ please change as required (and remove this comment)

    DefaultBufferSize := 1024.

    "Modified: / 07-06-2014 / 00:49:15 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!VDBInternalPipeStream class methodsFor:'instance creation'!

new
    ^ self newWithBufferSize: DefaultBufferSize

    "Modified: / 10-06-2014 / 00:26:32 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

newWithBufferSize: bufferSize
    ^ self basicNew initializeWithBufferSize: bufferSize

    "Created: / 07-06-2014 / 00:48:48 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!VDBInternalPipeStream methodsFor:'accessing'!

atEnd
    ^ closed and:[ last == 0 ]

    "Modified: / 07-06-2014 / 01:06:55 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

close
    closed := true.

    "Modified: / 07-06-2014 / 01:02:33 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

next
    "return the next element from the stream (might block until something is written)"

    | c |

    last == 0 ifTrue:[ 
        closed ifTrue:[ 
            ^ nil.
        ].
        dataAvailable wait.
    ].
    c := buffer at: first.
    first == last ifTrue:[ 
        first := 1.
        last := 0.
    ] ifFalse:[
       first ~~ buffer size ifTrue:[
           first := first + 1.
       ] ifFalse:[ 
           first := 1.
       ].
    ].
    spaceAvailable signalForAll.
    ^ c

    "Modified: / 07-06-2014 / 01:06:36 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

nextAvailableBytes:max into:out startingAt:offset
    | count |

    last == 0 ifTrue:[ ^ 0 ].

    last >= first ifTrue:[         
        count := max min: (last - first + 1).    
        out replaceFrom:offset to: offset + count - 1 with: buffer startingAt: first.
        first := first + count.
        first > last ifTrue:[ 
            first := 1. 
            last := 0.
        ].
        spaceAvailable signalForAll.
    ] ifFalse:[ 
        "/ Wrap around
        count := max.
        first + count <= buffer size ifTrue:[ 
            out replaceFrom:offset to: offset + count - 1 with: buffer startingAt: first.
            first := (first + count) \\ buffer size.
            spaceAvailable signalForAll.
        ] ifFalse:[ 
            | rem |

            count := max min: (buffer size - first) + last.
            rem := buffer size - first.
            out replaceFrom: offset to: offset + (buffer size - first) with: buffer startingAt: first.

            out replaceFrom: offset + (buffer size - first + 1) to:  offset + count with: buffer startingAt: 1.
            rem == last ifTrue:[ 
                first := 1.
                last := 0.
            ] ifFalse:[
                first := rem + 1.
            ].
        ].
    ].
    ^ count

    "Modified: / 10-06-2014 / 00:24:13 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

nextPut:aCharacter
    closed ifTrue:[ 
        self class writeErrorSignal signal:'Pipe stream closed'.
    ].
    
    last > first ifTrue:[ 
        last < buffer size ifTrue:[ 
            last := last + 1.
        ] ifFalse:[
            "/ wrapping around
            first == 1 ifTrue:[ 
                spaceAvailable wait. "/ wait for some space.
            ].
            last := 1.
        ]
    ] ifFalse:[ 
        (last ~~ 0 and:[(last + 1) = (first)]) ifTrue:[ 
            spaceAvailable wait. "/ wait for some space.                
        ].
        last == buffer size ifTrue:[ 
            last := 1.
        ] ifFalse:[
            last := last + 1.
        ].
    ].
    buffer at: last put: aCharacter.
    dataAvailable signalForAll.

    "Modified: / 07-06-2014 / 01:04:10 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

nextPutAll:aCollection startingAt:firstIndex to: lastIndex
    "Append the elements with index from firstIndex to lastIndex
     of the argument, aCollection,  onto the receiver."

    closed ifTrue:[ 
        self class writeErrorSignal signal:'Pipe stream closed'.
    ].

    (aCollection isString or:[ aCollection isByteArray ]) ifTrue:[ 
        self nextPutBytes:lastIndex - firstIndex + 1 from: aCollection startingAt: firstIndex
    ] ifFalse:[ 
        super nextPutAll:aCollection startingAt:firstIndex to: lastIndex
    ].

    "Created: / 09-06-2014 / 21:57:15 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

size
    last == 0 ifTrue:[ ^ 0 ].
    last >= first ifTrue:[ 
        ^ last - first + 1 
    ] ifFalse:[ 
        ^ buffer size - first + 1 + last
    ].

    "Modified: / 07-06-2014 / 01:08:53 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!VDBInternalPipeStream methodsFor:'initialization'!

initializeWithBufferSize: bufferSize
    buffer := String new: bufferSize.
    first := 1.
    last := 0.

    dataAvailable := Semaphore new.
    spaceAvailable := Semaphore new.

    closed := false

    "Created: / 07-06-2014 / 00:49:41 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!VDBInternalPipeStream methodsFor:'non homogenous writing'!

nextPutBytes:count from:anObject startingAt:start
    "Write count bytes from an object starting at index start.
     Return the number of bytes written.
     The object must have non-pointer indexed instvars 
     (i.e. be a ByteArray, String, Float- or DoubleArray).     
     Use with care - non object oriented i/o.
     This is provided for compatibility with externalStream;
     to support binary storage"

    closed ifTrue:[ 
        self class writeErrorSignal signal:'Pipe stream closed'.
    ].

    "/ Special case - empty buffer
    last == 0 ifTrue:[ 
        count <= buffer size ifTrue:[ 
            buffer replaceFrom:1 to: count with: anObject startingAt: start.
            last := count.
            dataAvailable signalForAll.
        ] ifFalse:[ 
            | rem |

            rem := count - buffer size.
            buffer replaceFrom:1 to: buffer size with: anObject startingAt: start.
            last := buffer size.
            dataAvailable signalForAll.
            self nextPutBytes:rem from:anObject startingAt:start + buffer size.
        ].
        ^ self.
    ].

    "/ Check if there's space...
    [ ((last + 1) \\ buffer size) == first ] whileTrue:[ 
        spaceAvailable wait.
    ].

    last >= first ifTrue:[ 
        last < buffer size ifTrue:[
            last + count <= buffer size ifTrue:[ 
                buffer replaceFrom:last + 1 to: last + count with: anObject startingAt: start.
                last := last + count.
                dataAvailable signalForAll.
            ] ifFalse:[
                | rem |

                rem := count - (buffer size - last).
                buffer replaceFrom:last + 1 to: buffer size with: anObject startingAt: start.
                last := buffer size.
                dataAvailable signalForAll.
                self nextPutBytes:rem from:anObject startingAt:start + rem.
            ].
        ] ifFalse:[
            "/ wrapping around
            first > count ifTrue:[ 
                buffer replaceFrom:1 to: count with: anObject startingAt: start.
                last := count.
                dataAvailable signalForAll.
            ] ifFalse:[ 
                | rem |    

                rem := count - first.
                buffer replaceFrom:1 to: first - 1 with: anObject startingAt: start.  
                last := first - 1.
                dataAvailable signalForAll.
                self nextPutBytes:rem from:anObject startingAt:start + rem.
            ].
        ].
    ] ifFalse:[ 
        last + count < first ifTrue:[ 
            buffer replaceFrom:last + 1 to: last + count with: anObject startingAt: start.
            last := last + count.
            dataAvailable signalForAll.
        ] ifFalse:[ 
            | rem |    

            rem := count - (first - last + 1).
            buffer replaceFrom: last to: last + rem with: anObject startingAt: start.  
            last := last + rem.
            dataAvailable signalForAll.
            self nextPutBytes:rem from:anObject startingAt:start + rem.    
        ].
    ]

    "Created: / 09-06-2014 / 22:04:15 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 09-06-2014 / 23:39:08 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!VDBInternalPipeStream methodsFor:'private'!

contentsSpecies
    ^ buffer class

    "Created: / 09-06-2014 / 21:50:45 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!VDBInternalPipeStream methodsFor:'synchronization'!

readWait
    last == 0 ifTrue:[ 
        dataAvailable wait.
    ].

    "Modified: / 07-06-2014 / 01:09:29 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!VDBInternalPipeStream methodsFor:'writing'!

nextPutAll:aCollection
    "Put all elements of the argument, aCollection onto the receiver."

    ^ self nextPutAll: aCollection startingAt: 1 to: aCollection size

    "Created: / 09-06-2014 / 21:58:13 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !


VDBInternalPipeStream initialize!
